@@ -2,9 +2,7 @@ import os
 import posixpath
 import sys
 import unittest
-from functools import partial
-from posixpath import realpath, abspath, dirname, basename, ALLOW_MISSING
-from test import support
+from posixpath import realpath, abspath, dirname, basename
 from test import test_genericpath
 from test.support import import_helper
 from test.support import os_helper
@@ -38,26 +36,6 @@ def safe_rmdir(dirname):
     except OSError:
         pass
 
-def _parameterize(*parameters):
-    """Simplistic decorator to parametrize a test
-
-    Runs the decorated test multiple times in subTest, with a value from
-    'parameters' passed as an extra positional argument.
-    Does *not* call doCleanups() after each run.
-
-    Not for general use. Intended to avoid indenting for easier backports.
-
-    See https://discuss.python.org/t/91827 for discussing generalizations.
-    """
-    def _parametrize_decorator(func):
-        def _parameterized(self, *args, **kwargs):
-            for parameter in parameters:
-                with self.subTest(parameter):
-                    func(self, *args, parameter, **kwargs)
-        return _parameterized
-    return _parametrize_decorator
-
-
 class PosixPathTest(unittest.TestCase):
 
     def setUp(self):
@@ -200,8 +178,6 @@ class PosixPathTest(unittest.TestCase):
     def test_ismount(self):
         self.assertIs(posixpath.ismount("/"), True)
         self.assertIs(posixpath.ismount(b"/"), True)
-        self.assertIs(posixpath.ismount(FakePath("/")), True)
-        self.assertIs(posixpath.ismount(FakePath(b"/")), True)
 
     def test_ismount_non_existent(self):
         # Non-existent mountpoint.
@@ -217,8 +193,7 @@ class PosixPathTest(unittest.TestCase):
         self.assertIs(posixpath.ismount('/\x00'), False)
         self.assertIs(posixpath.ismount(b'/\x00'), False)
 
-    @unittest.skipUnless(os_helper.can_symlink(),
-                         "Test requires symlink support")
+    @os_helper.skip_unless_symlink
     def test_ismount_symlinks(self):
         # Symlinks are never mountpoints.
         try:
@@ -328,62 +303,100 @@ class PosixPathTest(unittest.TestCase):
                 for path in ('~', '~/.local', '~vstinner/'):
                     self.assertEqual(posixpath.expanduser(path), path)
 
+    NORMPATH_CASES = [
+        ("", "."),
+        ("/", "/"),
+        ("/.", "/"),
+        ("/./", "/"),
+        ("/.//.", "/"),
+        ("/foo", "/foo"),
+        ("/foo/bar", "/foo/bar"),
+        ("//", "//"),
+        ("///", "/"),
+        ("///foo/.//bar//", "/foo/bar"),
+        ("///foo/.//bar//.//..//.//baz///", "/foo/baz"),
+        ("///..//./foo/.//bar", "/foo/bar"),
+        (".", "."),
+        (".//.", "."),
+        ("..", ".."),
+        ("../", ".."),
+        ("../foo", "../foo"),
+        ("../../foo", "../../foo"),
+        ("../foo/../bar", "../bar"),
+        ("../../foo/../bar/./baz/boom/..", "../../bar/baz"),
+        ("/..", "/"),
+        ("/..", "/"),
+        ("/../", "/"),
+        ("/..//", "/"),
+        ("//.", "//"),
+        ("//..", "//"),
+        ("//...", "//..."),
+        ("//../foo", "//foo"),
+        ("//../../foo", "//foo"),
+        ("/../foo", "/foo"),
+        ("/../../foo", "/foo"),
+        ("/../foo/../", "/"),
+        ("/../foo/../bar", "/bar"),
+        ("/../../foo/../bar/./baz/boom/..", "/bar/baz"),
+        ("/../../foo/../bar/./baz/boom/.", "/bar/baz/boom"),
+        ("foo/../bar/baz", "bar/baz"),
+        ("foo/../../bar/baz", "../bar/baz"),
+        ("foo/../../../bar/baz", "../../bar/baz"),
+        ("foo///../bar/.././../baz/boom", "../baz/boom"),
+        ("foo/bar/../..///../../baz/boom", "../../baz/boom"),
+        ("/foo/..", "/"),
+        ("/foo/../..", "/"),
+        ("//foo/..", "//"),
+        ("//foo/../..", "//"),
+        ("///foo/..", "/"),
+        ("///foo/../..", "/"),
+        ("////foo/..", "/"),
+        ("/////foo/..", "/"),
+    ]
+
     def test_normpath(self):
-        self.assertEqual(posixpath.normpath(""), ".")
-        self.assertEqual(posixpath.normpath("/"), "/")
-        self.assertEqual(posixpath.normpath("//"), "//")
-        self.assertEqual(posixpath.normpath("///"), "/")
-        self.assertEqual(posixpath.normpath("///foo/.//bar//"), "/foo/bar")
-        self.assertEqual(posixpath.normpath("///foo/.//bar//.//..//.//baz"),
-                         "/foo/baz")
-        self.assertEqual(posixpath.normpath("///..//./foo/.//bar"), "/foo/bar")
+        for path, expected in self.NORMPATH_CASES:
+            with self.subTest(path):
+                result = posixpath.normpath(path)
+                self.assertEqual(result, expected)
 
-        self.assertEqual(posixpath.normpath(b""), b".")
-        self.assertEqual(posixpath.normpath(b"/"), b"/")
-        self.assertEqual(posixpath.normpath(b"//"), b"//")
-        self.assertEqual(posixpath.normpath(b"///"), b"/")
-        self.assertEqual(posixpath.normpath(b"///foo/.//bar//"), b"/foo/bar")
-        self.assertEqual(posixpath.normpath(b"///foo/.//bar//.//..//.//baz"),
-                         b"/foo/baz")
-        self.assertEqual(posixpath.normpath(b"///..//./foo/.//bar"),
-                         b"/foo/bar")
+            path = path.encode('utf-8')
+            expected = expected.encode('utf-8')
+            with self.subTest(path, type=bytes):
+                result = posixpath.normpath(path)
+                self.assertEqual(result, expected)
 
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_curdir(self, kwargs):
-        self.assertEqual(realpath('.', **kwargs), os.getcwd())
-        self.assertEqual(realpath('./.', **kwargs), os.getcwd())
-        self.assertEqual(realpath('/'.join(['.'] * 100), **kwargs), os.getcwd())
+    def test_realpath_curdir(self):
+        self.assertEqual(realpath('.'), os.getcwd())
+        self.assertEqual(realpath('./.'), os.getcwd())
+        self.assertEqual(realpath('/'.join(['.'] * 100)), os.getcwd())
 
-        self.assertEqual(realpath(b'.', **kwargs), os.getcwdb())
-        self.assertEqual(realpath(b'./.', **kwargs), os.getcwdb())
-        self.assertEqual(realpath(b'/'.join([b'.'] * 100), **kwargs), os.getcwdb())
+        self.assertEqual(realpath(b'.'), os.getcwdb())
+        self.assertEqual(realpath(b'./.'), os.getcwdb())
+        self.assertEqual(realpath(b'/'.join([b'.'] * 100)), os.getcwdb())
 
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_pardir(self, kwargs):
-        self.assertEqual(realpath('..', **kwargs), dirname(os.getcwd()))
-        self.assertEqual(realpath('../..', **kwargs), dirname(dirname(os.getcwd())))
-        self.assertEqual(realpath('/'.join(['..'] * 100), **kwargs), '/')
+    def test_realpath_pardir(self):
+        self.assertEqual(realpath('..'), dirname(os.getcwd()))
+        self.assertEqual(realpath('../..'), dirname(dirname(os.getcwd())))
+        self.assertEqual(realpath('/'.join(['..'] * 100)), '/')
 
-        self.assertEqual(realpath(b'..', **kwargs), dirname(os.getcwdb()))
-        self.assertEqual(realpath(b'../..', **kwargs), dirname(dirname(os.getcwdb())))
-        self.assertEqual(realpath(b'/'.join([b'..'] * 100), **kwargs), b'/')
+        self.assertEqual(realpath(b'..'), dirname(os.getcwdb()))
+        self.assertEqual(realpath(b'../..'), dirname(dirname(os.getcwdb())))
+        self.assertEqual(realpath(b'/'.join([b'..'] * 100)), b'/')
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': ALLOW_MISSING})
-    def test_realpath_basic(self, kwargs):
+    def test_realpath_basic(self):
         # Basic operation.
         try:
             os.symlink(ABSTFN+"1", ABSTFN)
-            self.assertEqual(realpath(ABSTFN, **kwargs), ABSTFN+"1")
+            self.assertEqual(realpath(ABSTFN), ABSTFN+"1")
         finally:
             os_helper.unlink(ABSTFN)
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
     def test_realpath_strict(self):
         # Bug #43757: raise FileNotFoundError in strict mode if we encounter
@@ -395,116 +408,15 @@ class PosixPathTest(unittest.TestCase):
         finally:
             os_helper.unlink(ABSTFN)
 
-    def test_realpath_invalid_paths(self):
-        path = '/\x00'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(ValueError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-        path = b'/\x00'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(ValueError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-        path = '/nonexistent/x\x00'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-        path = b'/nonexistent/x\x00'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-        path = '/\x00/..'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(ValueError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-        path = b'/\x00/..'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(ValueError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-
-        path = '/nonexistent/x\x00/..'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-        path = b'/nonexistent/x\x00/..'
-        self.assertRaises(ValueError, realpath, path, strict=False)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-        self.assertRaises(ValueError, realpath, path, strict=ALLOW_MISSING)
-
-        path = '/\udfff'
-        if sys.platform == 'win32':
-            self.assertEqual(realpath(path, strict=False), path)
-            self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-            self.assertEqual(realpath(path, strict=ALLOW_MISSING), path)
-        else:
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=False)
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=True)
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=ALLOW_MISSING)
-        path = '/nonexistent/\udfff'
-        if sys.platform == 'win32':
-            self.assertEqual(realpath(path, strict=False), path)
-            self.assertEqual(realpath(path, strict=ALLOW_MISSING), path)
-        else:
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=False)
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=ALLOW_MISSING)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-        path = '/\udfff/..'
-        if sys.platform == 'win32':
-            self.assertEqual(realpath(path, strict=False), '/')
-            self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-            self.assertEqual(realpath(path, strict=ALLOW_MISSING), '/')
-        else:
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=False)
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=True)
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=ALLOW_MISSING)
-        path = '/nonexistent/\udfff/..'
-        if sys.platform == 'win32':
-            self.assertEqual(realpath(path, strict=False), '/nonexistent')
-            self.assertEqual(realpath(path, strict=ALLOW_MISSING), '/nonexistent')
-        else:
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=False)
-            self.assertRaises(UnicodeEncodeError, realpath, path, strict=ALLOW_MISSING)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-
-        path = b'/\xff'
-        if sys.platform == 'win32':
-            self.assertRaises(UnicodeDecodeError, realpath, path, strict=False)
-            self.assertRaises(UnicodeDecodeError, realpath, path, strict=True)
-            self.assertRaises(UnicodeDecodeError, realpath, path, strict=ALLOW_MISSING)
-        else:
-            self.assertEqual(realpath(path, strict=False), path)
-            self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-            self.assertEqual(realpath(path, strict=ALLOW_MISSING), path)
-        path = b'/nonexistent/\xff'
-        if sys.platform == 'win32':
-            self.assertRaises(UnicodeDecodeError, realpath, path, strict=False)
-            self.assertRaises(UnicodeDecodeError, realpath, path, strict=ALLOW_MISSING)
-        else:
-            self.assertEqual(realpath(path, strict=False), path)
-        self.assertRaises(FileNotFoundError, realpath, path, strict=True)
-
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': ALLOW_MISSING})
-    def test_realpath_relative(self, kwargs):
+    def test_realpath_relative(self):
         try:
             os.symlink(posixpath.relpath(ABSTFN+"1"), ABSTFN)
-            self.assertEqual(realpath(ABSTFN, **kwargs), ABSTFN+"1")
+            self.assertEqual(realpath(ABSTFN), ABSTFN+"1")
         finally:
             os_helper.unlink(ABSTFN)
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
-    @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': ALLOW_MISSING})
-    def test_realpath_missing_pardir(self, kwargs):
-        try:
-            os.symlink(os_helper.TESTFN + "1", os_helper.TESTFN)
-            self.assertEqual(
-                realpath("nonexistent/../" + os_helper.TESTFN, **kwargs), ABSTFN + "1")
-        finally:
-            os_helper.unlink(os_helper.TESTFN)
-
     @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
     def test_realpath_symlink_loops(self):
@@ -546,41 +458,39 @@ class PosixPathTest(unittest.TestCase):
             os_helper.unlink(ABSTFN+"c")
             os_helper.unlink(ABSTFN+"a")
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_symlink_loops_strict(self, kwargs):
+    def test_realpath_symlink_loops_strict(self):
         # Bug #43757, raise OSError if we get into an infinite symlink loop in
-        # the strict modes.
+        # strict mode.
         try:
             os.symlink(ABSTFN, ABSTFN)
-            self.assertRaises(OSError, realpath, ABSTFN, **kwargs)
+            self.assertRaises(OSError, realpath, ABSTFN, strict=True)
 
             os.symlink(ABSTFN+"1", ABSTFN+"2")
             os.symlink(ABSTFN+"2", ABSTFN+"1")
-            self.assertRaises(OSError, realpath, ABSTFN+"1", **kwargs)
-            self.assertRaises(OSError, realpath, ABSTFN+"2", **kwargs)
+            self.assertRaises(OSError, realpath, ABSTFN+"1", strict=True)
+            self.assertRaises(OSError, realpath, ABSTFN+"2", strict=True)
 
-            self.assertRaises(OSError, realpath, ABSTFN+"1/x", **kwargs)
-            self.assertRaises(OSError, realpath, ABSTFN+"1/..", **kwargs)
-            self.assertRaises(OSError, realpath, ABSTFN+"1/../x", **kwargs)
+            self.assertRaises(OSError, realpath, ABSTFN+"1/x", strict=True)
+            self.assertRaises(OSError, realpath, ABSTFN+"1/..", strict=True)
+            self.assertRaises(OSError, realpath, ABSTFN+"1/../x", strict=True)
             os.symlink(ABSTFN+"x", ABSTFN+"y")
             self.assertRaises(OSError, realpath,
-                              ABSTFN+"1/../" + basename(ABSTFN) + "y", **kwargs)
+                              ABSTFN+"1/../" + basename(ABSTFN) + "y", strict=True)
             self.assertRaises(OSError, realpath,
-                              ABSTFN+"1/../" + basename(ABSTFN) + "1", **kwargs)
+                              ABSTFN+"1/../" + basename(ABSTFN) + "1", strict=True)
 
             os.symlink(basename(ABSTFN) + "a/b", ABSTFN+"a")
-            self.assertRaises(OSError, realpath, ABSTFN+"a", **kwargs)
+            self.assertRaises(OSError, realpath, ABSTFN+"a", strict=True)
 
             os.symlink("../" + basename(dirname(ABSTFN)) + "/" +
                        basename(ABSTFN) + "c", ABSTFN+"c")
-            self.assertRaises(OSError, realpath, ABSTFN+"c", **kwargs)
+            self.assertRaises(OSError, realpath, ABSTFN+"c", strict=True)
 
             # Test using relative path as well.
             with os_helper.change_cwd(dirname(ABSTFN)):
-                self.assertRaises(OSError, realpath, basename(ABSTFN), **kwargs)
+                self.assertRaises(OSError, realpath, basename(ABSTFN), strict=True)
         finally:
             os_helper.unlink(ABSTFN)
             os_helper.unlink(ABSTFN+"1")
@@ -589,34 +499,30 @@ class PosixPathTest(unittest.TestCase):
             os_helper.unlink(ABSTFN+"c")
             os_helper.unlink(ABSTFN+"a")
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_repeated_indirect_symlinks(self, kwargs):
+    def test_realpath_repeated_indirect_symlinks(self):
         # Issue #6975.
         try:
             os.mkdir(ABSTFN)
             os.symlink('../' + basename(ABSTFN), ABSTFN + '/self')
             os.symlink('self/self/self', ABSTFN + '/link')
-            self.assertEqual(realpath(ABSTFN + '/link', **kwargs), ABSTFN)
+            self.assertEqual(realpath(ABSTFN + '/link'), ABSTFN)
         finally:
             os_helper.unlink(ABSTFN + '/self')
             os_helper.unlink(ABSTFN + '/link')
             safe_rmdir(ABSTFN)
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_deep_recursion(self, kwargs):
+    def test_realpath_deep_recursion(self):
         depth = 10
         try:
             os.mkdir(ABSTFN)
             for i in range(depth):
                 os.symlink('/'.join(['%d' % i] * 10), ABSTFN + '/%d' % (i + 1))
             os.symlink('.', ABSTFN + '/0')
-            self.assertEqual(realpath(ABSTFN + '/%d' % depth, **kwargs), ABSTFN)
+            self.assertEqual(realpath(ABSTFN + '/%d' % depth), ABSTFN)
 
             # Test using relative path as well.
             with os_helper.change_cwd(ABSTFN):
@@ -626,11 +532,9 @@ class PosixPathTest(unittest.TestCase):
                 os_helper.unlink(ABSTFN + '/%d' % i)
             safe_rmdir(ABSTFN)
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': ALLOW_MISSING})
-    def test_realpath_resolve_parents(self, kwargs):
+    def test_realpath_resolve_parents(self):
         # We also need to resolve any symlinks in the parents of a relative
         # path passed to realpath. E.g.: current working directory is
         # /usr/doc with 'doc' being a symlink to /usr/share/doc. We call
@@ -641,18 +545,15 @@ class PosixPathTest(unittest.TestCase):
             os.symlink(ABSTFN + "/y", ABSTFN + "/k")
 
             with os_helper.change_cwd(ABSTFN + "/k"):
-                self.assertEqual(realpath("a", **kwargs),
-                                    ABSTFN + "/y/a")
+                self.assertEqual(realpath("a"), ABSTFN + "/y/a")
         finally:
             os_helper.unlink(ABSTFN + "/k")
             safe_rmdir(ABSTFN + "/y")
             safe_rmdir(ABSTFN)
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_resolve_before_normalizing(self, kwargs):
+    def test_realpath_resolve_before_normalizing(self):
         # Bug #990669: Symbolic links should be resolved before we
         # normalize the path. E.g.: if we have directories 'a', 'k' and 'y'
         # in the following hierarchy:
@@ -667,10 +568,10 @@ class PosixPathTest(unittest.TestCase):
             os.symlink(ABSTFN + "/k/y", ABSTFN + "/link-y")
 
             # Absolute path.
-            self.assertEqual(realpath(ABSTFN + "/link-y/..", **kwargs), ABSTFN + "/k")
+            self.assertEqual(realpath(ABSTFN + "/link-y/.."), ABSTFN + "/k")
             # Relative path.
             with os_helper.change_cwd(dirname(ABSTFN)):
-                self.assertEqual(realpath(basename(ABSTFN) + "/link-y/..", **kwargs),
+                self.assertEqual(realpath(basename(ABSTFN) + "/link-y/.."),
                                  ABSTFN + "/k")
         finally:
             os_helper.unlink(ABSTFN + "/link-y")
@@ -678,11 +579,9 @@ class PosixPathTest(unittest.TestCase):
             safe_rmdir(ABSTFN + "/k")
             safe_rmdir(ABSTFN)
 
-    @unittest.skipUnless(hasattr(os, "symlink"),
-                         "Missing symlink implementation")
+    @os_helper.skip_unless_symlink
     @skip_if_ABSTFN_contains_backslash
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_resolve_first(self, kwargs):
+    def test_realpath_resolve_first(self):
         # Bug #1213894: The first component of the path, if not absolute,
         # must be resolved too.
 
@@ -692,70 +591,13 @@ class PosixPathTest(unittest.TestCase):
             os.symlink(ABSTFN, ABSTFN + "link")
             with os_helper.change_cwd(dirname(ABSTFN)):
                 base = basename(ABSTFN)
-                self.assertEqual(realpath(base + "link", **kwargs), ABSTFN)
-                self.assertEqual(realpath(base + "link/k", **kwargs), ABSTFN + "/k")
+                self.assertEqual(realpath(base + "link"), ABSTFN)
+                self.assertEqual(realpath(base + "link/k"), ABSTFN + "/k")
         finally:
             os_helper.unlink(ABSTFN + "link")
             safe_rmdir(ABSTFN + "/k")
             safe_rmdir(ABSTFN)
 
-    @os_helper.skip_unless_symlink
-    @skip_if_ABSTFN_contains_backslash
-    @unittest.skipIf(os.chmod not in os.supports_follow_symlinks, "Can't set symlink permissions")
-    @unittest.skipIf(sys.platform != "darwin", "only macOS requires read permission to readlink()")
-    @_parameterize({'strict': True}, {'strict': ALLOW_MISSING})
-    def test_realpath_unreadable_symlink_strict(self, kwargs):
-        try:
-            os.symlink(ABSTFN+"1", ABSTFN)
-            os.chmod(ABSTFN, 0o000, follow_symlinks=False)
-            with self.assertRaises(PermissionError):
-                realpath(ABSTFN, **kwargs)
-            with self.assertRaises(PermissionError):
-                realpath(ABSTFN + '/foo', **kwargs),
-            with self.assertRaises(PermissionError):
-                realpath(ABSTFN + '/../foo', **kwargs)
-            with self.assertRaises(PermissionError):
-                realpath(ABSTFN + '/foo/..', **kwargs)
-        finally:
-            os.chmod(ABSTFN, 0o755, follow_symlinks=False)
-            os.unlink(ABSTFN)
-
-    @skip_if_ABSTFN_contains_backslash
-    @os_helper.skip_unless_symlink
-    def test_realpath_unreadable_directory(self):
-        try:
-            os.mkdir(ABSTFN)
-            os.mkdir(ABSTFN + '/k')
-            os.chmod(ABSTFN, 0o000)
-            self.assertEqual(realpath(ABSTFN, strict=False), ABSTFN)
-            self.assertEqual(realpath(ABSTFN, strict=True), ABSTFN)
-            self.assertEqual(realpath(ABSTFN, strict=ALLOW_MISSING), ABSTFN)
-
-            try:
-                os.stat(ABSTFN)
-            except PermissionError:
-                pass
-            else:
-                self.skipTest('Cannot block permissions')
-
-            self.assertEqual(realpath(ABSTFN + '/k', strict=False),
-                             ABSTFN + '/k')
-            self.assertRaises(PermissionError, realpath, ABSTFN + '/k',
-                              strict=True)
-            self.assertRaises(PermissionError, realpath, ABSTFN + '/k',
-                              strict=ALLOW_MISSING)
-
-            self.assertEqual(realpath(ABSTFN + '/missing', strict=False),
-                             ABSTFN + '/missing')
-            self.assertRaises(PermissionError, realpath, ABSTFN + '/missing',
-                              strict=True)
-            self.assertRaises(PermissionError, realpath, ABSTFN + '/missing',
-                              strict=ALLOW_MISSING)
-        finally:
-            os.chmod(ABSTFN, 0o755)
-            safe_rmdir(ABSTFN + '/k')
-            safe_rmdir(ABSTFN)
-
     def test_relpath(self):
         (real_getcwd, os.getcwd) = (os.getcwd, lambda: r"/home/user/bar")
         try:
@@ -932,12 +774,9 @@ class PathLikeTests(unittest.TestCase):
     def test_path_abspath(self):
         self.assertPathEqual(self.path.abspath)
 
-    @_parameterize({}, {'strict': True}, {'strict': ALLOW_MISSING})
-    def test_path_realpath(self, kwargs):
+    def test_path_realpath(self):
         self.assertPathEqual(self.path.realpath)
 
-        self.assertPathEqual(partial(self.path.realpath, **kwargs))
-
     def test_path_relpath(self):
         self.assertPathEqual(self.path.relpath)
 
